@@ -58,6 +58,7 @@
 	!defined(CONFIG_HAVE_FENTRY) || \
 	!defined(CONFIG_MODULES) || \
 	!defined(CONFIG_SYSFS) || \
+	!defined(CONFIG_STACKTRACE) || \
 	!defined(CONFIG_KALLSYMS_ALL)
 #error "CONFIG_FUNCTION_TRACER, CONFIG_HAVE_FENTRY, CONFIG_MODULES, CONFIG_SYSFS, CONFIG_KALLSYMS_ALL kernel config options are required"
 #endif
@@ -82,6 +83,11 @@ struct kpatch_kallsyms_args {
 	unsigned long pos;
 };
 
+struct kpatch_apply_patch_args {
+	struct kpatch_module *kpmod;
+	bool replace;
+};
+
 /* this is a double loop, use goto instead of break */
 #define do_for_each_linked_func(kpmod, func) {				\
 	struct kpatch_object *_object;					\
@@ -139,7 +145,7 @@ static int (*kpatch_set_memory_ro)(unsigned long addr, int numpages);
 
 #define MAX_STACK_TRACE_DEPTH   64
 static unsigned long stack_entries[MAX_STACK_TRACE_DEPTH];
-struct stack_trace trace = {
+static struct stack_trace trace = {
 	.max_entries	= ARRAY_SIZE(stack_entries),
 	.entries	= &stack_entries[0],
 };
@@ -206,7 +212,8 @@ static inline int kpatch_compare_addresses(unsigned long stack_addr,
 }
 
 static int kpatch_backtrace_address_verify(struct kpatch_module *kpmod,
-					   unsigned long address)
+					   unsigned long address,
+					   bool replace)
 {
 	struct kpatch_func *func;
 	int i;
@@ -241,8 +248,11 @@ static int kpatch_backtrace_address_verify(struct kpatch_module *kpmod,
 	} while_for_each_linked_func();
 
 	/* in the replace case, need to check the func hash as well */
-	hash_for_each_rcu(kpatch_func_hash, i, func, node) {
-		if (func->op == KPATCH_OP_UNPATCH && !func->force) {
+	if (replace) {
+		hash_for_each_rcu(kpatch_func_hash, i, func, node) {
+			if (func->op != KPATCH_OP_UNPATCH || func->force)
+				continue;
+
 			ret = kpatch_compare_addresses(address,
 						       func->new_addr,
 						       func->new_size,
@@ -261,7 +271,8 @@ static int kpatch_backtrace_address_verify(struct kpatch_module *kpmod,
  *
  * This function is called from stop_machine() context.
  */
-static int kpatch_verify_activeness_safety(struct kpatch_module *kpmod)
+static int kpatch_verify_activeness_safety(struct kpatch_module *kpmod,
+					   bool replace)
 {
 	struct task_struct *g, *t;
 	int i;
@@ -283,7 +294,8 @@ static int kpatch_verify_activeness_safety(struct kpatch_module *kpmod)
 			if (trace.entries[i] == ULONG_MAX)
 				break;
 			ret = kpatch_backtrace_address_verify(kpmod,
-							      trace.entries[i]);
+							      trace.entries[i],
+							      replace);
 			if (ret)
 				goto out;
 		}
@@ -305,16 +317,61 @@ out:
 	return ret;
 }
 
+static inline int pre_patch_callback(struct kpatch_object *object)
+{
+	int ret;
+
+	if (kpatch_object_linked(object) &&
+	    object->pre_patch_callback) {
+		ret = (*object->pre_patch_callback)(object);
+		if (ret) {
+			object->callbacks_enabled = false;
+			return ret;
+		}
+	}
+	object->callbacks_enabled = true;
+
+	return 0;
+}
+
+static inline void post_patch_callback(struct kpatch_object *object)
+{
+	if (kpatch_object_linked(object) &&
+	    object->post_patch_callback &&
+	    object->callbacks_enabled)
+		(*object->post_patch_callback)(object);
+}
+
+static inline void pre_unpatch_callback(struct kpatch_object *object)
+{
+	if (kpatch_object_linked(object) &&
+	    object->pre_unpatch_callback &&
+	    object->callbacks_enabled)
+		(*object->pre_unpatch_callback)(object);
+}
+
+static inline void post_unpatch_callback(struct kpatch_object *object)
+{
+	if (kpatch_object_linked(object) &&
+	    object->post_unpatch_callback &&
+	    object->callbacks_enabled)
+		(*object->post_unpatch_callback)(object);
+}
+
 /* Called from stop_machine */
 static int kpatch_apply_patch(void *data)
 {
-	struct kpatch_module *kpmod = data;
+	struct kpatch_apply_patch_args *args = data;
+	struct kpatch_module *kpmod;
 	struct kpatch_func *func;
-	struct kpatch_hook *hook;
+	struct hlist_node *tmp;
 	struct kpatch_object *object;
 	int ret;
+	int i;
 
-	ret = kpatch_verify_activeness_safety(kpmod);
+	kpmod = args->kpmod;
+
+	ret = kpatch_verify_activeness_safety(kpmod, args->replace);
 	if (ret) {
 		kpatch_state_finish(KPATCH_STATE_FAILURE);
 		return ret;
@@ -344,14 +401,22 @@ static int kpatch_apply_patch(void *data)
 		return -EBUSY;
 	}
 
-	/* run any user-defined load hooks */
-	list_for_each_entry(object, &kpmod->objects, list) {
-		if (!kpatch_object_linked(object))
-			continue;
-		list_for_each_entry(hook, &object->hooks_load, list)
-			(*hook->hook)();
+	/*
+	 * The new patch has been applied successfully. Remove the functions
+	 * provided by the replaced patches (if any) from hash, to make sure
+	 * they will not be executed anymore.
+	 */
+	if (args->replace) {
+		hash_for_each_safe(kpatch_func_hash, i, tmp, func, node) {
+			if (func->op != KPATCH_OP_UNPATCH)
+				continue;
+			hash_del_rcu(&func->node);
+		}
 	}
 
+	/* run any user-defined post-patch callbacks */
+	list_for_each_entry(object, &kpmod->objects, list)
+		post_patch_callback(object);
 
 	return 0;
 }
@@ -361,35 +426,39 @@ static int kpatch_remove_patch(void *data)
 {
 	struct kpatch_module *kpmod = data;
 	struct kpatch_func *func;
-	struct kpatch_hook *hook;
 	struct kpatch_object *object;
 	int ret;
 
-	ret = kpatch_verify_activeness_safety(kpmod);
+	ret = kpatch_verify_activeness_safety(kpmod, false);
 	if (ret) {
 		kpatch_state_finish(KPATCH_STATE_FAILURE);
 		return ret;
 	}
 
+	/* run any user-defined pre-unpatch callbacks */
+	list_for_each_entry(object, &kpmod->objects, list)
+		pre_unpatch_callback(object);
+
 	/* Check if any inconsistent NMI has happened while updating */
 	ret = kpatch_state_finish(KPATCH_STATE_SUCCESS);
-	if (ret == KPATCH_STATE_FAILURE)
-		return -EBUSY;
+	if (ret == KPATCH_STATE_FAILURE) {
+		ret = -EBUSY;
+		goto err;
+	}
 
 	/* Succeeded, remove all updating funcs from hash table */
 	do_for_each_linked_func(kpmod, func) {
 		hash_del_rcu(&func->node);
 	} while_for_each_linked_func();
 
-	/* run any user-defined unload hooks */
-	list_for_each_entry(object, &kpmod->objects, list) {
-		if (!kpatch_object_linked(object))
-			continue;
-		list_for_each_entry(hook, &object->hooks_unload, list)
-			(*hook->hook)();
-	}
-
 	return 0;
+
+err:
+	/* undo pre-unpatch callbacks by calling post-patch counterparts */
+	list_for_each_entry(object, &kpmod->objects, list)
+		post_patch_callback(object);
+
+	return ret;
 }
 
 /*
@@ -590,7 +659,11 @@ static int kpatch_find_external_symbol(const char *objname, const char *name,
 	sym = find_symbol(name, NULL, NULL, true, true);
 	preempt_enable();
 	if (sym) {
+#ifdef CONFIG_HAVE_ARCH_PREL32_RELOCATIONS
+		*addr = (unsigned long)offset_to_ptr(&sym->value_offset);
+#else
 		*addr = sym->value;
+#endif
 		return 0;
 	}
 
@@ -635,6 +708,7 @@ static int kpatch_write_relocations(struct kpatch_module *kpmod,
 		case R_X86_64_NONE:
 			continue;
 		case R_X86_64_PC32:
+		case R_X86_64_PLT32:
 			loc = dynrela->dest;
 			val = (u32)(dynrela->src + dynrela->addend -
 				    dynrela->dest);
@@ -647,7 +721,7 @@ static int kpatch_write_relocations(struct kpatch_module *kpmod,
 			break;
 		case R_X86_64_64:
 			loc = dynrela->dest;
-			val = dynrela->src;
+			val = dynrela->src + dynrela->addend;
 			size = 8;
 			break;
 		default:
@@ -723,8 +797,10 @@ static int kpatch_unlink_object(struct kpatch_object *object)
 		}
 	}
 
-	if (object->mod)
+	if (object->mod) {
 		module_put(object->mod);
+		object->mod = NULL;
+	}
 
 	return 0;
 }
@@ -801,14 +877,13 @@ err_put:
 	return ret;
 }
 
-static int kpatch_module_notify(struct notifier_block *nb, unsigned long action,
-				void *data)
+static int kpatch_module_notify_coming(struct notifier_block *nb,
+				       unsigned long action, void *data)
 {
 	struct module *mod = data;
 	struct kpatch_module *kpmod;
 	struct kpatch_object *object;
 	struct kpatch_func *func;
-	struct kpatch_hook *hook;
 	int ret = 0;
 	bool found = false;
 
@@ -839,29 +914,110 @@ done:
 
 	pr_notice("patching newly loaded module '%s'\n", object->name);
 
-	/* run any user-defined load hooks */
-	list_for_each_entry(hook, &object->hooks_load, list)
-		(*hook->hook)();
+	/* run user-defined pre-patch callback */
+	ret = pre_patch_callback(object);
+	if (ret) {
+		pr_err("pre-patch callback failed!\n");
+		goto out;	/* and WARN */
+	}
 
 	/* add to the global func hash */
 	list_for_each_entry(func, &object->funcs, list)
 		hash_add_rcu(kpatch_func_hash, &func->node, func->old_addr);
 
+	/* run user-defined post-patch callback */
+	post_patch_callback(object);
 out:
 	up(&kpatch_mutex);
 
 	/* no way to stop the module load on error */
 	WARN(ret, "error (%d) patching newly loaded module '%s'\n", ret,
 	     object->name);
+
 	return 0;
 }
 
+static int kpatch_module_notify_going(struct notifier_block *nb,
+				      unsigned long action, void *data)
+{
+	struct module *mod = data;
+	struct kpatch_module *kpmod;
+	struct kpatch_object *object;
+	struct kpatch_func *func;
+	bool found = false;
+
+	if (action != MODULE_STATE_GOING)
+		return 0;
+
+	down(&kpatch_mutex);
+
+	list_for_each_entry(kpmod, &kpmod_list, list) {
+		list_for_each_entry(object, &kpmod->objects, list) {
+			if (!kpatch_object_linked(object))
+				continue;
+			if (!strcmp(object->name, mod->name)) {
+				found = true;
+				goto done;
+			}
+		}
+	}
+done:
+	if (!found)
+		goto out;
+
+	/* run user-defined pre-unpatch callback */
+	pre_unpatch_callback(object);
+
+	/* remove from the global func hash */
+	list_for_each_entry(func, &object->funcs, list)
+		hash_del_rcu(&func->node);
+
+	/* run user-defined pre-unpatch callback */
+	post_unpatch_callback(object);
+
+	kpatch_unlink_object(object);
+
+out:
+	up(&kpatch_mutex);
+
+	return 0;
+}
+
+/*
+ * Remove the obsolete functions from the ftrace filter.
+ * Return 1 if one or more of such functions have 'force' flag set,
+ * 0 otherwise.
+ */
+static int kpatch_ftrace_remove_unpatched_funcs(void)
+{
+	struct kpatch_module *kpmod;
+	struct kpatch_func *func;
+	int force = 0;
+
+	list_for_each_entry(kpmod, &kpmod_list, list) {
+		do_for_each_linked_func(kpmod, func) {
+			if (func->op != KPATCH_OP_UNPATCH)
+				continue;
+			if (func->force)
+				force = 1;
+			WARN_ON(kpatch_ftrace_remove_func(func->old_addr));
+		} while_for_each_linked_func();
+	}
+
+	return force;
+}
+
 int kpatch_register(struct kpatch_module *kpmod, bool replace)
 {
-	int ret, i, force = 0;
+	int ret, i;
 	struct kpatch_object *object, *object_err = NULL;
 	struct kpatch_func *func;
 
+	struct kpatch_apply_patch_args args = {
+		.kpmod = kpmod,
+		.replace = replace,
+	};
+
 	if (!kpmod->mod || list_empty(&kpmod->objects))
 		return -EINVAL;
 
@@ -909,28 +1065,41 @@ int kpatch_register(struct kpatch_module *kpmod, bool replace)
 
 	kpatch_state_updating();
 
-	/*
-	 * Idle the CPUs, verify activeness safety, and atomically make the new
-	 * functions visible to the ftrace handler.
-	 */
-	ret = stop_machine(kpatch_apply_patch, kpmod, NULL);
+	/* run any user-defined pre-patch callbacks */
+	list_for_each_entry(object, &kpmod->objects, list) {
+		ret = pre_patch_callback(object);
+		if(ret){
+			pr_err("pre-patch callback failed!\n");
+			kpatch_state_finish(KPATCH_STATE_FAILURE);
+			break;
+		}
+	}
+
+	/* if pre_patch_callback succeed. */
+	if (!ret) {
+		/*
+		 * Idle the CPUs, verify activeness safety, and atomically make the new
+		 * functions visible to the ftrace handler.
+		 */
+		ret = stop_machine(kpatch_apply_patch, &args, NULL);
+	}
+
+	/* if pre_patch_callback or stop_machine failed */
+	if (ret) {
+		list_for_each_entry(object, &kpmod->objects, list)
+			post_unpatch_callback(object);
+	}
 
 	/*
-	 * For the replace case, remove any obsolete funcs from the hash and
-	 * the ftrace filter, and disable the owning patch module so that it
-	 * can be removed.
+	 * For the replace case, remove any obsolete funcs from the ftrace
+	 * filter, and disable the owning patch module so that it can be
+	 * removed.
 	 */
 	if (!ret && replace) {
 		struct kpatch_module *kpmod2, *safe;
+		int force;
 
-		hash_for_each_rcu(kpatch_func_hash, i, func, node) {
-			if (func->op != KPATCH_OP_UNPATCH)
-				continue;
-			if (func->force)
-				force = 1;
-			hash_del_rcu(&func->node);
-			WARN_ON(kpatch_ftrace_remove_func(func->old_addr));
-		}
+		force = kpatch_ftrace_remove_unpatched_funcs();
 
 		list_for_each_entry_safe(kpmod2, safe, &kpmod_list, list) {
 			if (kpmod == kpmod2)
@@ -1053,6 +1222,11 @@ int kpatch_unregister(struct kpatch_module *kpmod)
 
 	ret = stop_machine(kpatch_remove_patch, kpmod, NULL);
 
+	if (!ret) {
+		/* run any user-defined post-unpatch callbacks */
+		list_for_each_entry(object, &kpmod->objects, list)
+			post_unpatch_callback(object);
+	}
 	/* NMI handlers can return to normal now */
 	kpatch_state_idle();
 
@@ -1100,10 +1274,14 @@ out:
 EXPORT_SYMBOL(kpatch_unregister);
 
 
-static struct notifier_block kpatch_module_nb = {
-	.notifier_call = kpatch_module_notify,
+static struct notifier_block kpatch_module_nb_coming = {
+	.notifier_call = kpatch_module_notify_coming,
 	.priority = INT_MIN, /* called last */
 };
+static struct notifier_block kpatch_module_nb_going = {
+	.notifier_call = kpatch_module_notify_going,
+	.priority = INT_MAX, /* called first */
+};
 
 static int kpatch_init(void)
 {
@@ -1125,12 +1303,17 @@ static int kpatch_init(void)
 	if (!kpatch_root_kobj)
 		return -ENOMEM;
 
-	ret = register_module_notifier(&kpatch_module_nb);
+	ret = register_module_notifier(&kpatch_module_nb_coming);
 	if (ret)
 		goto err_root_kobj;
+	ret = register_module_notifier(&kpatch_module_nb_going);
+	if (ret)
+		goto err_unregister_coming;
 
 	return 0;
 
+err_unregister_coming:
+	WARN_ON(unregister_module_notifier(&kpatch_module_nb_coming));
 err_root_kobj:
 	kobject_put(kpatch_root_kobj);
 	return ret;
@@ -1141,7 +1324,8 @@ static void kpatch_exit(void)
 	rcu_barrier();
 
 	WARN_ON(kpatch_num_patched != 0);
-	WARN_ON(unregister_module_notifier(&kpatch_module_nb));
+	WARN_ON(unregister_module_notifier(&kpatch_module_nb_coming));
+	WARN_ON(unregister_module_notifier(&kpatch_module_nb_going));
 	kobject_put(kpatch_root_kobj);
 }
 
