@@ -65,6 +65,18 @@ bool is_text_section(struct section *sec)
 		(sec->sh.sh_flags & SHF_EXECINSTR));
 }
 
+bool is_patchable_function_entries_section(struct section *sec)
+{
+	char *name;
+
+	if (is_rela_section(sec))
+		name = sec->base->name;
+	else
+		name = sec->name;
+
+	return !strcmp(name, "__patchable_function_entries");
+}
+
 bool is_debug_section(struct section *sec)
 {
 	char *name;
@@ -142,6 +154,8 @@ unsigned int absolute_rela_type(struct kpatch_elf *kelf)
 		return R_X86_64_64;
 	case S390:
 		return R_390_64;
+	case AARCH64:
+		return R_AARCH64_ABS64;
 	default:
 		ERROR("unsupported arch");
 	}
@@ -206,6 +220,7 @@ long rela_target_offset(struct kpatch_elf *kelf, struct section *relasec,
 
 	switch(kelf->arch) {
 	case PPC64:
+	case AARCH64:
 		add_off = 0;
 		break;
 	case X86_64:
@@ -255,6 +270,8 @@ unsigned int insn_length(struct kpatch_elf *kelf, void *addr)
 
 	switch(kelf->arch) {
 
+	case AARCH64:
+		return 4;
 	case X86_64:
 		insn_init(&decoded_insn, addr, 1);
 		insn_get_length(&decoded_insn);
@@ -559,6 +576,7 @@ struct kpatch_elf *kpatch_elf_open(const char *name)
 	int fd;
 	struct kpatch_elf *kelf;
 	struct section *relasec;
+	struct section *pfesec;
 	GElf_Ehdr ehdr;
 
 	fd = open(name, O_RDONLY);
@@ -593,6 +611,9 @@ struct kpatch_elf *kpatch_elf_open(const char *name)
 	case EM_S390:
 		kelf->arch = S390;
 		break;
+	case EM_AARCH64:
+		kelf->arch = AARCH64;
+		break;
 	default:
 		ERROR("Unsupported target architecture");
 	}
@@ -614,9 +635,17 @@ struct kpatch_elf *kpatch_elf_open(const char *name)
 	 * These sections aren't used by ftrace on this arch, so do not
 	 * bother reading/writing them for x86_64.
 	 */
-	if (kelf->arch != X86_64)
-		if (find_section_by_name(&kelf->sections, "__patchable_function_entries"))
+	if (kelf->arch != X86_64) {
+		pfesec = find_section_by_name(&kelf->sections, "__patchable_function_entries");
+		if (pfesec) {
 			kelf->has_pfe = true;
+			/*
+			 * Check if the pfe section generated by the toolchain has the SHF_LINK_ORDER
+			 * flag set. This info is used while generating pfe sections in the diff object.
+			 */
+			kelf->pfe_ordered = !!(pfesec->sh.sh_flags & SHF_LINK_ORDER);
+		}
+	}
 
 	return kelf;
 }
@@ -887,19 +916,12 @@ void kpatch_create_symtab(struct kpatch_elf *kelf)
 	symtab->sh.sh_info = nr_local;
 }
 
-struct section *create_section_pair(struct kpatch_elf *kelf, char *name,
-                                    int entsize, int nr)
+struct section *create_section(struct kpatch_elf *kelf, char *name,
+                               int entsize, int nr)
 {
-	char *relaname;
-	struct section *sec, *relasec;
+	struct section *sec;
 	int size = entsize * nr;
 
-	relaname = malloc(strlen(name) + strlen(".rela") + 1);
-	if (!relaname)
-		ERROR("malloc");
-	strcpy(relaname, ".rela");
-	strcat(relaname, name);
-
 	/* allocate text section resources */
 	ALLOC_LINK(sec, &kelf->sections);
 	sec->name = name;
@@ -922,6 +944,24 @@ struct section *create_section_pair(struct kpatch_elf *kelf, char *name,
 	sec->sh.sh_flags = SHF_ALLOC;
 	sec->sh.sh_size = size;
 
+	return sec;
+
+}
+
+struct section *create_section_pair(struct kpatch_elf *kelf, char *name,
+                                    int entsize, int nr)
+{
+	char *relaname;
+	struct section *sec, *relasec;
+
+	relaname = malloc(strlen(name) + strlen(".rela") + 1);
+	if (!relaname)
+		ERROR("malloc");
+	strcpy(relaname, ".rela");
+	strcat(relaname, name);
+
+	sec = create_section(kelf, name, entsize, nr);
+
 	/* allocate rela section resources */
 	ALLOC_LINK(relasec, &kelf->sections);
 	relasec->name = relaname;
